from .testing_util import *

from qastle import *

import ast


def assert_identical_atom(initial_text):
    initial_python_ast = ast.parse(initial_text)
    final_text_ast = python_ast_to_text_ast(initial_python_ast)
    assert final_text_ast == initial_text.strip(' \t\r\n')
    final_python_ast = text_ast_to_python_ast(initial_text)
    assert_ast_nodes_are_equal(final_python_ast, initial_python_ast)


def assert_equivalent_literal(initial_text):
    initial_python_ast = ast.parse(initial_text)
    final_text_ast = python_ast_to_text_ast(initial_python_ast)
    assert ast.literal_eval(final_text_ast) == ast.literal_eval(initial_text)
    final_python_ast = text_ast_to_python_ast(initial_text)
    assert_ast_nodes_are_equal(final_python_ast, initial_python_ast)


def assert_equivalent_python_ast_and_text_ast(initial_python_ast, initial_text_ast):
    final_text_ast = python_ast_to_text_ast(initial_python_ast)
    assert final_text_ast == initial_text_ast
    final_python_ast = text_ast_to_python_ast(initial_text_ast)
    assert_ast_nodes_are_equal(final_python_ast, initial_python_ast)


def assert_equivalent_python_text_and_text_ast(python_text, initial_text_ast):
    initial_python_ast = ast.parse(python_text)
    assert_equivalent_python_ast_and_text_ast(initial_python_ast, initial_text_ast)


def test_empty():
    assert_identical_atom('')


def test_whitespace():
    assert_identical_atom(' \t\r\n')


def test_identifiers():
    assert_identical_atom('xyz')


def test_string_literals():
    assert_identical_atom("''")
    assert_identical_atom("'as\"df'")
    assert_identical_atom('"as\'df"')


def test_numeric_literals():
    assert_identical_atom('0')
    assert_identical_atom('1.2')
    assert_identical_atom('3e+21')
    assert_identical_atom('4e-22')
    assert_identical_atom('5.6e+23')
    assert_identical_atom('7.8e-24')

    assert_equivalent_literal('1.')
    assert_equivalent_literal('.2')
    assert_equivalent_literal('3.e4')
    assert_equivalent_literal('5.e+6')
    assert_equivalent_literal('7.e-8')
    assert_equivalent_literal('.9e10')
    assert_equivalent_literal('.11e+12')
    assert_equivalent_literal('.13e-14')


def test_list():
    assert_equivalent_python_text_and_text_ast('[]', '(list)')
    assert_equivalent_python_text_and_text_ast('[0]', '(list 0)')
    assert_equivalent_python_text_and_text_ast('[0, 1, 2]', '(list 0 1 2)')


def test_tuple():
    assert python_source_to_text_ast('()') == '(list)'
    assert python_source_to_text_ast('(0,)') == '(list 0)'
    assert python_source_to_text_ast('(0, 1, 2)') == '(list 0 1 2)'


def test_dict():
    assert_equivalent_python_text_and_text_ast('{}', '(dict (list) (list))')
    assert_equivalent_python_text_and_text_ast('{0: 0}', '(dict (list 0) (list 0))')
    assert_equivalent_python_text_and_text_ast("{0: 0, 1: 'a', 'b': 2, 'abc': 'abc'}",
                                               "(dict (list 0 1 'b' 'abc') (list 0 'a' 2 'abc'))")


def test_attr():
    assert_equivalent_python_text_and_text_ast('a.b', "(attr a 'b')")


def test_subscript():
    assert_equivalent_python_text_and_text_ast('a[0]', '(subscript a 0)')
    assert_equivalent_python_text_and_text_ast("a['b']", "(subscript a 'b')")


def test_call():
    assert_equivalent_python_text_and_text_ast('a()', '(call a)')
    assert_equivalent_python_text_and_text_ast('a(0, 1, 2)', '(call a 0 1 2)')


def test_if():
    assert_equivalent_python_text_and_text_ast('a if b else c', '(if b a c)')


def test_unary_operators():
    assert python_source_to_text_ast('+1') == '1'
    assert_ast_nodes_are_equal(text_ast_to_python_ast('+1'), ast.parse('1'))
    assert_equivalent_python_text_and_text_ast('+a', '(+ a)')
    assert_equivalent_python_text_and_text_ast('-1', '-1')
    assert_equivalent_python_text_and_text_ast('-a', '(- a)')
    assert_equivalent_python_text_and_text_ast('not True', '(not True)')
    assert_equivalent_python_text_and_text_ast('~1', '(~ 1)')


def test_binary_operators():
    assert_equivalent_python_text_and_text_ast('1 + 2', '(+ 1 2)')
    assert_equivalent_python_text_and_text_ast('1 - 2', '(- 1 2)')
    assert_equivalent_python_text_and_text_ast('1 * 2', '(* 1 2)')
    assert_equivalent_python_text_and_text_ast('1 / 2', '(/ 1 2)')
    assert_equivalent_python_text_and_text_ast('1 % 2', '(% 1 2)')
    assert_equivalent_python_text_and_text_ast('1 ** 2', '(** 1 2)')
    assert_equivalent_python_text_and_text_ast('1 // 2', '(// 1 2)')
    assert_equivalent_python_text_and_text_ast('1 & 2', '(& 1 2)')
    assert_equivalent_python_text_and_text_ast('1 | 2', '(| 1 2)')
    assert_equivalent_python_text_and_text_ast('1 ^ 2', '(^ 1 2)')
    assert_equivalent_python_text_and_text_ast('1 << 2', '(<< 1 2)')
    assert_equivalent_python_text_and_text_ast('1 >> 2', '(>> 1 2)')


def test_boolean_operators():
    assert_equivalent_python_text_and_text_ast('True and False', '(and True False)')
    assert_equivalent_python_text_and_text_ast('True or False', '(or True False)')
    assert python_source_to_text_ast('a and b and c') == '(and (and a b) c)'


def test_comparison_operators():
    assert_equivalent_python_text_and_text_ast('1 == 2', '(== 1 2)')
    assert_equivalent_python_text_and_text_ast('1 != 2', '(!= 1 2)')
    assert_equivalent_python_text_and_text_ast('1 < 2', '(< 1 2)')
    assert_equivalent_python_text_and_text_ast('1 <= 2', '(<= 1 2)')
    assert_equivalent_python_text_and_text_ast('1 > 2', '(> 1 2)')
    assert_equivalent_python_text_and_text_ast('1 >= 2', '(>= 1 2)')
    assert python_source_to_text_ast('1 < 2 < 3') == '(and (< 1 2) (< 2 3))'
    assert python_source_to_text_ast('1 < 2 < 3 < 4') == '(and (and (< 1 2) (< 2 3)) (< 3 4))'


def test_lambda():
    assert_equivalent_python_text_and_text_ast('lambda: 0', '(lambda (list) 0)')
    assert_equivalent_python_text_and_text_ast('lambda x, y, z: x', '(lambda (list x y z) x)')


def test_Where():
    where_node = Where(source=unwrap_ast(ast.parse('data_source')),
                       predicate=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(where_node),
                                              '(Where data_source (lambda (list e) e))')


def test_Select():
    select_node = Select(source=unwrap_ast(ast.parse('data_source')),
                         selector=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(select_node),
                                              '(Select data_source (lambda (list e) e))')


def test_SelectMany():
    selectmany_node = SelectMany(source=unwrap_ast(ast.parse('data_source')),
                                 selector=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(selectmany_node),
                                              '(SelectMany data_source (lambda (list e) e))')


def test_First():
    first_node = First(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(First data_source)')


def test_Last():
    first_node = Last(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Last data_source)')


def test_ElementAt():
    first_node = ElementAt(source=unwrap_ast(ast.parse('data_source')),
                           index=unwrap_ast(ast.parse('2')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(ElementAt data_source 2)')


def test_Contains():
    first_node = Contains(source=unwrap_ast(ast.parse('data_source')),
                          value=unwrap_ast(ast.parse('element')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Contains data_source element)')


def test_Aggregate():
    aggregate_node = Aggregate(source=unwrap_ast(ast.parse('data_source')),
                               seed=unwrap_ast(ast.parse('0')),
                               func=unwrap_ast(ast.parse('lambda v, e : v + e')))
    text_ast = '(Aggregate data_source 0 (lambda (list v e) (+ v e)))'
    assert_equivalent_python_ast_and_text_ast(wrap_ast(aggregate_node),
                                              text_ast)


def test_Count():
    first_node = Count(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Count data_source)')


def test_Max():
    first_node = Max(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Max data_source)')


def test_Min():
    first_node = Min(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Min data_source)')


def test_Sum():
    first_node = Sum(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Sum data_source)')


def test_All():
    select_node = All(source=unwrap_ast(ast.parse('data_source')),
                      predicate=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(select_node),
                                              '(All data_source (lambda (list e) e))')


def test_Any():
    select_node = Any(source=unwrap_ast(ast.parse('data_source')),
                      predicate=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(select_node),
                                              '(Any data_source (lambda (list e) e))')


def test_Concat():
    first_node = Concat(first=unwrap_ast(ast.parse('sequence1')),
                        second=unwrap_ast(ast.parse('sequence2')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Concat sequence1 sequence2)')


def test_Zip():
    first_node = Zip(source=unwrap_ast(ast.parse('data_source')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(first_node),
                                              '(Zip data_source)')


def test_OrderBy():
    orderby_node = OrderBy(source=unwrap_ast(ast.parse('data_source')),
                           key_selector=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(orderby_node),
                                              '(OrderBy data_source (lambda (list e) e))')


def test_OrderByDescending():
    orderbydescending_node = OrderByDescending(source=unwrap_ast(ast.parse('data_source')),
                                               key_selector=unwrap_ast(ast.parse('lambda e: e')))
    assert_equivalent_python_ast_and_text_ast(
        wrap_ast(orderbydescending_node),
        '(OrderByDescending data_source (lambda (list e) e))')


def test_Choose():
    choose_node = Choose(source=unwrap_ast(ast.parse('data_source')),
                         n=unwrap_ast(ast.parse('2')))
    assert_equivalent_python_ast_and_text_ast(wrap_ast(choose_node),
                                              '(Choose data_source 2)')



from .transform import PythonASTToTextASTTransformer, TextASTToPythonASTTransformer
from .parse import parse

import ast


def python_source_to_python_ast(python_source):
    return ast.parse(python_source)


def python_source_to_text_ast(python_source):
    python_ast = python_source_to_python_ast(python_source)
    return python_ast_to_text_ast(python_ast)


def python_ast_to_text_ast(python_ast):
    return PythonASTToTextASTTransformer().visit(python_ast)


def text_ast_to_python_ast(text_ast):
    tree = parse(text_ast)
    return TextASTToPythonASTTransformer().transform(tree)



import ast


def are_ast_nodes_equal(ast_1, ast_2):
    return ast.dump(ast_1) == ast.dump(ast_2)


def assert_ast_nodes_are_equal(ast_1, ast_2):
    assert are_ast_nodes_equal(ast_1, ast_2), ('ast_1 = ' + ast.dump(ast_1)
                                               + ', ast_2 = ' + ast.dump(ast_2))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='qastle',
                 version='0.16.1',
                 description='Query AST Language Expressions',
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 packages=setuptools.find_packages(exclude=['tests']),
                 python_requires=('>=2.7, '
                                  '!=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <3.12'),
                 install_requires=['lark'],
                 extras_require={'test': ['flake8', 'pytest', 'pytest-cov']},
                 package_data={'qastle': ['syntax.lark']},
                 author='Mason Proffitt',
                 author_email='masonlp@uw.edu',
                 url='https://github.com/iris-hep/qastle')



from .linq_util import (Where, Select, SelectMany, First, Last, ElementAt, Contains, Aggregate,
                        Count, Max, Min, Sum, All, Any, Concat, Zip, OrderBy, OrderByDescending,
                        Choose)
from .ast_util import wrap_ast, unwrap_ast

import lark

import ast
import sys


UnaryOp_ops = {'+':   ast.UAdd,
               '-':   ast.USub,
               'not': ast.Not,
               '~':   ast.Invert}

BinOp_ops = {'+': ast.Add,
             '-': ast.Sub,
             '*': ast.Mult,
             '/': ast.Div,
             '%': ast.Mod,
             '**': ast.Pow,
             '//': ast.FloorDiv,
             '&': ast.BitAnd,
             '|': ast.BitOr,
             '^': ast.BitXor,
             '<<': ast.LShift,
             '>>': ast.RShift}

BoolOp_ops = {'and': ast.And,
              'or':  ast.Or}

Compare_ops = {'==': ast.Eq,
               '!=': ast.NotEq,
               '<':  ast.Lt,
               '<=': ast.LtE,
               '>':  ast.Gt,
               '>=': ast.GtE}

op_strings = {value: key
              for dictionary in [UnaryOp_ops, BinOp_ops, BoolOp_ops, Compare_ops]
              for key, value in dictionary.items()}

flexible_ops = ('+', '-')


class PythonASTToTextASTTransformer(ast.NodeVisitor):
    def visit_Module(self, node):
        n_children = len(node.body)
        if n_children == 0:
            return ''
        elif n_children == 1:
            return self.visit(node.body[0])
        else:
            raise SyntaxError('A record must contain zero or one expressions; found '
                              + str(n_children))

    def visit_Expr(self, node):
        return self.visit(node.value)

    def visit_Name(self, node):
        return node.id

    def visit_Constant(self, node):
        return repr(node.value)

    def visit_Num(self, node):
        return repr(node.n)

    def visit_Str(self, node):
        return repr(node.s)

    def visit_NameConstant(self, node):
        return repr(node.value)

    @staticmethod
    def make_composite_node_string(node_type, *fields):
        return '(' + node_type + ''.join([' ' + field for field in fields]) + ')'

    def visit_List(self, node):
        return self.make_composite_node_string('list',
                                               *[self.visit(element) for element in node.elts])

    def visit_Tuple(self, node):
        return self.visit_List(node)

    def visit_Dict(self, node):
        return self.make_composite_node_string('dict',
                                               self.visit(ast.List(elts=node.keys)),
                                               self.visit(ast.List(elts=node.values)))

    def visit_Attribute(self, node):
        return self.make_composite_node_string('attr', self.visit(node.value), repr(node.attr))

    def visit_Subscript(self, node):
        return self.make_composite_node_string('subscript',
                                               self.visit(node.value),
                                               self.visit(node.slice))

    def visit_Index(self, node):
        return self.visit(node.value)

    def visit_Call(self, node):
        return self.make_composite_node_string('call',
                                               self.visit(node.func),
                                               *[self.visit(arg) for arg in node.args])

    def visit_IfExp(self, node):
        return self.make_composite_node_string('if',
                                               self.visit(node.test),
                                               self.visit(node.body),
                                               self.visit(node.orelse))

    def visit_UnaryOp(self, node):
        if (hasattr(ast, 'Constant') and isinstance(node.operand, ast.Constant)
           or isinstance(node.operand, ast.Num)):
            if isinstance(node.op, ast.UAdd):
                return self.visit(node.operand)
            elif isinstance(node.op, ast.USub):
                return self.visit(ast.Num(n=-node.operand.n))
        return self.make_composite_node_string(op_strings[type(node.op)],
                                               self.visit(node.operand))

    def visit_BinOp(self, node):
        return self.make_composite_node_string(op_strings[type(node.op)],
                                               self.visit(node.left),
                                               self.visit(node.right))

    def visit_BoolOp(self, node):
        if len(node.values) < 2:
            raise SyntaxError('Boolean operator must have at least 2 operands; found: '
                              + str(len(node.values)))
        rep = self.visit(node.values[0])
        for value in node.values[1:]:
            rep = self.make_composite_node_string(op_strings[type(node.op)],
                                                  rep,
                                                  self.visit(value))
        return rep

    def visit_Compare(self, node):
        if len(node.ops) < 1:
            raise SyntaxError('Compare node must have at least 1 operation; found: '
                              + str(len(node.ops)))
        left = self.visit(node.left)
        right = self.visit(node.comparators[0])
        rep = self.make_composite_node_string(op_strings[type(node.ops[0])],
                                              left,
                                              right)
        for operator, comparator in zip(node.ops[1:], node.comparators[1:]):
            left = right
            right = self.visit(comparator)
            new_comparison = self.make_composite_node_string(op_strings[type(operator)],
                                                             left,
                                                             right)
            rep = self.make_composite_node_string('and', rep, new_comparison)
        return rep

    def visit_Lambda(self, node):
        return self.make_composite_node_string('lambda',
                                               self.visit(node.args),
                                               self.visit(node.body))

    def visit_arguments(self, node):
        return self.visit(ast.List(elts=node.args))

    def visit_arg(self, node):
        return node.arg

    def visit_Where(self, node):
        return self.make_composite_node_string('Where',
                                               self.visit(node.source),
                                               self.visit(node.predicate))

    def visit_Select(self, node):
        return self.make_composite_node_string('Select',
                                               self.visit(node.source),
                                               self.visit(node.selector))

    def visit_SelectMany(self, node):
        return self.make_composite_node_string('SelectMany',
                                               self.visit(node.source),
                                               self.visit(node.selector))

    def visit_First(self, node):
        return self.make_composite_node_string('First', self.visit(node.source))

    def visit_Last(self, node):
        return self.make_composite_node_string('Last', self.visit(node.source))

    def visit_ElementAt(self, node):
        return self.make_composite_node_string('ElementAt',
                                               self.visit(node.source),
                                               self.visit(node.index))

    def visit_Contains(self, node):
        return self.make_composite_node_string('Contains',
                                               self.visit(node.source),
                                               self.visit(node.value))

    def visit_Aggregate(self, node):
        return self.make_composite_node_string('Aggregate',
                                               self.visit(node.source),
                                               self.visit(node.seed),
                                               self.visit(node.func))

    def visit_Count(self, node):
        return self.make_composite_node_string('Count', self.visit(node.source))

    def visit_Max(self, node):
        return self.make_composite_node_string('Max', self.visit(node.source))

    def visit_Min(self, node):
        return self.make_composite_node_string('Min', self.visit(node.source))

    def visit_Sum(self, node):
        return self.make_composite_node_string('Sum', self.visit(node.source))

    def visit_All(self, node):
        return self.make_composite_node_string('All',
                                               self.visit(node.source),
                                               self.visit(node.predicate))

    def visit_Any(self, node):
        return self.make_composite_node_string('Any',
                                               self.visit(node.source),
                                               self.visit(node.predicate))

    def visit_Concat(self, node):
        return self.make_composite_node_string('Concat',
                                               self.visit(node.first),
                                               self.visit(node.second))

    def visit_Zip(self, node):
        return self.make_composite_node_string('Zip', self.visit(node.source))

    def visit_OrderBy(self, node):
        return self.make_composite_node_string('OrderBy',
                                               self.visit(node.source),
                                               self.visit(node.key_selector))

    def visit_OrderByDescending(self, node):
        return self.make_composite_node_string('OrderByDescending',
                                               self.visit(node.source),
                                               self.visit(node.key_selector))

    def visit_Choose(self, node):
        return self.make_composite_node_string('Choose',
                                               self.visit(node.source),
                                               self.visit(node.n))

    def generic_visit(self, node):
        raise SyntaxError('Unsupported node type: ' + str(type(node)))


class TextASTToPythonASTTransformer(lark.Transformer):
    def record(self, children):
        if (len(children) == 0
           or isinstance(children[0], lark.Token) and children[0].type == 'WHITESPACE'):
            return wrap_ast()
        else:
            return wrap_ast(children[0])

    def expression(self, children):
        for child in children:
            if not (isinstance(child, lark.Token) and child.type == 'WHITESPACE'):
                return child
        raise SyntaxError('Expression does not contain a node')

    def atom(self, children):
        child = children[0]
        if child.type == 'NUMERIC_LITERAL':
            if child.value[0] == '+':
                child.value = child.value[1:]
        return unwrap_ast(ast.parse(child.value))

    def composite(self, children):
        fields = []
        for child in children:
            if isinstance(child, lark.Token):
                if child.type == 'NODE_TYPE':
                    node_type = child.value
                else:
                    pass
            elif isinstance(child, ast.AST):
                fields.append(child)
            else:
                pass

        if node_type == 'list':
            return ast.List(elts=fields, ctx=ast.Load())

        if node_type == 'dict':
            if len(fields) != 2:
                raise SyntaxError('Dictionary node must have two fields; found '
                                  + str(len(fields)))
            for field_index in range(2):
                if not isinstance(fields[field_index], ast.List):
                    raise SyntaxError('Dictionary fields must be lists; found '
                                      + str(type(fields[field_index])))
            return ast.Dict(keys=fields[0].elts, values=fields[1].elts)

        elif node_type == 'attr':
            if len(fields) != 2:
                raise SyntaxError('Attribute node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Str):
                raise SyntaxError('Attribute name must be a string; found ' + str(type(fields[1])))
            return ast.Attribute(value=fields[0], attr=fields[1].s, ctx=ast.Load())

        elif node_type == 'subscript':
            if len(fields) != 2:
                raise SyntaxError('Subscript node must have two fields; found ' + str(len(fields)))
            if sys.version_info[0] < 3 or (sys.version_info[0] == 3 and sys.version_info[1] < 9):
                return ast.Subscript(value=fields[0],
                                     slice=ast.Index(value=fields[1]),
                                     ctx=ast.Load())
            else:
                return ast.Subscript(value=fields[0], slice=fields[1], ctx=ast.Load())

        elif node_type == 'call':
            if len(fields) < 1:
                raise SyntaxError('Call node must have at least one field; found '
                                  + str(len(fields)))
            if sys.version_info[0] < 3:
                return ast.Call(func=fields[0],
                                args=fields[1:],
                                keywords=[],
                                starargs=None,
                                kwargs=None)
            else:
                return ast.Call(func=fields[0], args=fields[1:], keywords=[])

        elif node_type == 'if':
            if len(fields) != 3:
                raise SyntaxError('If node must have three fields; found '
                                  + str(len(fields)))
            return ast.IfExp(test=fields[0], body=fields[1], orelse=fields[2])

        elif node_type in UnaryOp_ops:
            if len(fields) == 1:
                return ast.UnaryOp(op=UnaryOp_ops[node_type](), operand=fields[0])
            elif node_type not in flexible_ops:
                raise SyntaxError(UnaryOp_ops[node_type].__name__
                                  + ' operator only supported for one operand; found '
                                  + str(len(fields)))

        if node_type in BinOp_ops:
            if len(fields) == 2:
                return ast.BinOp(left=fields[0], op=BinOp_ops[node_type](), right=fields[1])
            else:
                raise SyntaxError(BinOp_ops[node_type].__name__
                                  + ' operator only supported for two operands; found '
                                  + str(len(fields)))

        elif node_type in BoolOp_ops:
            if len(fields) == 2:
                return ast.BoolOp(op=BoolOp_ops[node_type](), values=fields)
            else:
                raise SyntaxError(BoolOp_ops[node_type].__name__
                                  + ' operator only supported for two operands; found '
                                  + str(len(fields)))

        elif node_type in Compare_ops:
            if len(fields) == 2:
                return ast.Compare(left=fields[0],
                                   ops=[Compare_ops[node_type]()],
                                   comparators=[fields[1]])
            else:
                raise SyntaxError(Compare_ops[node_type].__name__
                                  + ' operator only supported for two operands; found '
                                  + str(len(fields)))

        elif node_type == 'lambda':
            if len(fields) != 2:
                raise SyntaxError('Lambda node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[0], ast.List):
                raise SyntaxError('Lambda arguments must be in a list; found '
                                  + str(type(fields[0])))
            for arg in fields[0].elts:
                if not isinstance(arg, ast.Name):
                    raise SyntaxError('Lambda arguments must variable names; found '
                                      + str(type(arg)))
            if sys.version_info[0] < 3:
                return ast.Lambda(args=ast.arguments(args=[ast.Name(id=name.id, ctx=ast.Param())
                                                           for name in fields[0].elts],
                                                     vararg=None,
                                                     kwarg=None,
                                                     defaults=[]),
                                  body=fields[1])
            elif sys.version_info[0] == 3 and sys.version_info[1] < 8:
                return ast.Lambda(args=ast.arguments(args=[ast.arg(arg=name.id, annotation=None)
                                                           for name in fields[0].elts],
                                                     vararg=None,
                                                     kwonlyargs=[],
                                                     kw_defaults=[],
                                                     kwarg=None,
                                                     defaults=[]),
                                  body=fields[1])
            else:
                return ast.Lambda(args=ast.arguments(posonlyargs=[],
                                                     args=[ast.arg(arg=name.id,
                                                                   annotation=None,
                                                                   type_comment=None)
                                                           for name in fields[0].elts],
                                                     vararg=None,
                                                     kwonlyargs=[],
                                                     kw_defaults=[],
                                                     kwarg=None,
                                                     defaults=[]),
                                  body=fields[1])

        elif node_type == 'Where':
            if len(fields) != 2:
                raise SyntaxError('Where node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('Where predicate must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('Where predicate must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return Where(source=fields[0], predicate=fields[1])

        elif node_type == 'Select':
            if len(fields) != 2:
                raise SyntaxError('Select node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('Select selector must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('Select selector must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return Select(source=fields[0], selector=fields[1])

        elif node_type == 'SelectMany':
            if len(fields) != 2:
                raise SyntaxError('SelectMany node must have two fields; found '
                                  + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('SelectMany selector must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('SelectMany selector must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return SelectMany(source=fields[0], selector=fields[1])

        elif node_type == 'First':
            if len(fields) != 1:
                raise SyntaxError('First node must have one field; found ' + str(len(fields)))
            return First(source=fields[0])

        elif node_type == 'Last':
            if len(fields) != 1:
                raise SyntaxError('Last node must have one field; found ' + str(len(fields)))
            return Last(source=fields[0])

        elif node_type == 'ElementAt':
            if len(fields) != 2:
                raise SyntaxError('ElementAt node must have two fields; found ' + str(len(fields)))
            return ElementAt(source=fields[0], index=fields[1])

        elif node_type == 'Contains':
            if len(fields) != 2:
                raise SyntaxError('Contains node must have two fields; found ' + str(len(fields)))
            return Contains(source=fields[0], value=fields[1])

        elif node_type == 'Aggregate':
            if len(fields) != 3:
                raise SyntaxError('Aggregate node must have three fields; found '
                                  + str(len(fields)))
            if not isinstance(fields[2], ast.Lambda):
                raise SyntaxError('Aggregate func must be a lambda; found ' + str(type(fields[1])))
            if len(fields[2].args.args) != 2:
                raise SyntaxError('Aggregate func must have exactly two arguments; found '
                                  + str(len(fields[2].args.args)))
            return Aggregate(source=fields[0], seed=fields[1], func=fields[2])

        elif node_type == 'Count':
            if len(fields) != 1:
                raise SyntaxError('Count node must have one field; found ' + str(len(fields)))
            return Count(source=fields[0])

        elif node_type == 'Max':
            if len(fields) != 1:
                raise SyntaxError('Max node must have one field; found ' + str(len(fields)))
            return Max(source=fields[0])

        elif node_type == 'Min':
            if len(fields) != 1:
                raise SyntaxError('Min node must have one field; found ' + str(len(fields)))
            return Min(source=fields[0])

        elif node_type == 'Sum':
            if len(fields) != 1:
                raise SyntaxError('Sum node must have one field; found ' + str(len(fields)))
            return Sum(source=fields[0])

        elif node_type == 'All':
            if len(fields) != 2:
                raise SyntaxError('All node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('All predicate must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('All predicate must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return All(source=fields[0], predicate=fields[1])

        elif node_type == 'Any':
            if len(fields) != 2:
                raise SyntaxError('Any node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('Any predicate must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('Any predicate must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return Any(source=fields[0], predicate=fields[1])

        elif node_type == 'Concat':
            if len(fields) != 2:
                raise SyntaxError('Concat node must have two fields; found ' + str(len(fields)))
            return Concat(first=fields[0], second=fields[1])

        elif node_type == 'Zip':
            if len(fields) != 1:
                raise SyntaxError('Zip node must have one field; found ' + str(len(fields)))
            return Zip(source=fields[0])

        elif node_type == 'OrderBy':
            if len(fields) != 2:
                raise SyntaxError('OrderBy node must have two fields; found ' + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('OrderBy key selector must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('OrderBy key selector must have exactly one argument; found '
                                  + str(len(fields[1].args.args)))
            return OrderBy(source=fields[0], key_selector=fields[1])

        elif node_type == 'OrderByDescending':
            if len(fields) != 2:
                raise SyntaxError('OrderByDescending node must have two fields; found '
                                  + str(len(fields)))
            if not isinstance(fields[1], ast.Lambda):
                raise SyntaxError('OrderByDescending key selector must be a lambda; found '
                                  + str(type(fields[1])))
            if len(fields[1].args.args) != 1:
                raise SyntaxError('OrderByDescending key selector must have exactly one argument;'
                                  + ' found ' + str(len(fields[1].args.args)))
            return OrderByDescending(source=fields[0], key_selector=fields[1])

        elif node_type == 'Choose':
            if len(fields) != 2:
                raise SyntaxError('Choose node must have two fields; found ' + str(len(fields)))
            return Choose(source=fields[0], n=fields[1])

        else:
            raise SyntaxError('Unknown composite node type: ' + node_type)
